open Format
open Options
open Sig

module type S = sig
  type t

  exception Inconsistent

  val empty : t
  val normal_form :  Atom.t ->  Atom.t 
  val add : Atom.t -> t -> t
  val assume : Atom.t -> Formula.Set.t -> t -> t
  val query : Atom.t -> t -> bool
  val explain: Atom.t -> t -> Formula.Set.t
  val classof : t -> Term.t -> Term.t list
end

(*module CombineX (X1 : Sig.THEORY)(X2 : Sig.THEORY) = struct *)

module rec CombineX : Sig.THEORY = struct
  module MT = Term.Map
  let name = "[" ^ X1.name ^ " " ^ X2.name ^ "]" 

  type t = X1 of X1.t | X2 of X2.t | Empty of Term.t

  let is_mine_t _ = true 

  let is_mine_r = function
    | X1 r1 -> X1.is_mine_r r1
    | X2 r2 -> X2.is_mine_r r2
    | Empty t -> Sig.No t
	
  let is_an_eq a = match Atom.view a with Atom.Built _ -> false | _ -> true
  let is_mine_a a = X1.is_mine_a a || X2.is_mine_a a || is_an_eq a 
  let normal_form a =
    if X1.is_mine_a a then X1.normal_form a 
    else if X2.is_mine_a a then X2.normal_form a else a
      
  let compare a b = 
    match a , b with
      | X1 x , X1 y -> X1.compare x y
      | X2 x , X2 y -> X2.compare x y
      | Empty x , Empty y -> Term.compare x y
      | X1 _ , (X2 _ | Empty _)-> -1
      | X2 _ , X1 _ -> 1
      | X2 _, Empty _ -> -1
      | Empty _, (X1 _ | X2 _) -> 1
	  
  exception Unsolvable
    
  let print fmt = function
      X1 t -> (*fprintf fmt "%a" X1.print t*)
	fprintf fmt "X1(%s):%a" X1.name X1.print t
    | X2 t -> (*fprintf fmt "%a" X2.print t*)
	fprintf fmt "X2(%s):%a" X2.name X2.print t
    | Empty t -> (*fprintf fmt "%a" Term.print t*)
	fprintf fmt "Empty:%a" Term.print t
	
  let make_repr repr = 
    let repr1 x = 
      match repr x with 
	  X1 v -> v 
	| Empty t -> X1.make t
	| _ -> raise Not_found in
    let repr2 x = 
      match repr x with 
	  X2 v -> v 
	| Empty t -> X2.make t
	| _ -> raise Not_found in
    repr1, repr2
      
(*  let make t =
    if X1.is_mine_t t then X1 (X1.make t) else 
      if X2.is_mine_t t then X2 (X2.make t) else Empty t *)

  let make t =
    let rec make_rec t =
      let r1 = X1.make t in
      let r2 = X2.make t in
      match X1.is_mine_r r1 , X2.is_mine_r r2 with
	  Yes , No _ -> X1 r1
	| No _ , Yes -> X2 r2 
	| No t1 , No t2 ->
	    (match Term.equal t1 t , Term.equal t2 t with
		false , false -> assert false
	      | true , false -> make_rec t2
	      | false, true -> make_rec t1
	      | _ ,_ -> Empty t)
	| Yes , Yes -> assert false
    in make_rec t

  let color1 r1 = match X1.is_mine_r r1 with Yes -> X1 r1 | No t -> make t
  let color2 r2 = match X2.is_mine_r r2 with Yes -> X2 r2 | No t -> make t
  let color r = 
    match r with
	X1 r1 -> (match X1.is_mine_r r1 with Yes -> r | No t -> make t)
      | X2 r2 -> (match X2.is_mine_r r2 with Yes -> r | No t -> make t)
      | Empty t -> make t
	    
  let leaves = function 
      X1 t -> X1.leaves t 
    | X2 t -> X2.leaves t | Empty t -> Term.Set.singleton t

  let subst p v r = 
    let r' = match v , r with
	X1 v1 , X1 t1 ->  X1 (X1.subst p v1 t1)
      | X2 v2 , X2 t2 -> X2 (X2.subst p v2 t2)
      | _, Empty t0 -> if Term.compare p t0 = 0 then v else r
      | Empty v0, X1 t1 -> X1 (X1.subst p (X1.make v0) t1)
      | Empty v0, X2 t2 -> X2 (X2.subst p (X2.make v0) t2)
      | X1 v1, X2 t2 ->
	  begin
	    match X1.is_mine_r v1 with
	      | Yes -> r
	      | No v0 ->  X2 (X2.subst p (X2.make v0) t2)
	  end
      | X2 v2, X1 t1 ->
	  begin
	    match X2.is_mine_r v2 with
	      | Yes -> r
	      | No v0 ->  X1 (X1.subst p (X1.make v0) t1)
	  end
    in 
    color r'

  let solve1 x y = 
    List.map (fun (t,v) -> make t, X1 v) (X1.solve (X1.make x) (X1.make y))
  let solve2 x y = 
    List.map (fun (t,v) -> make t, X2 v) (X2.solve (X2.make x) (X2.make y))

  let add_mt p v mt = MT.add p (v::(try MT.find p mt with Not_found -> [])) mt
  let rec solve_list solved l =
    List.fold_left
      (fun solved -> function
	 | Empty x, (Empty y as b) -> 
	     if Term.equal x y then solved else add_mt x b solved
	 | X1 x, X1 y -> 
	     List.fold_left
	       (fun solved (p,v) -> 
		  match X1.is_mine_r v with
		    | No t when X2.is_mine_t p && X2.is_mine_t t -> 
			solve_list solved (solve2 p t)
		    | _ -> add_mt p (color1 v) solved)
	       solved (X1.solve x y)
	 | Empty x, X1 y | X1 y, Empty x -> 
	     List.fold_left
	       (fun solved (p,v) ->
		  match X1.is_mine_r v with
		    | No t when X2.is_mine_t p && X2.is_mine_t t ->
			solve_list solved (solve2 p t)
		    | _ -> add_mt p (color1 v) solved)
	       solved (X1.solve (X1.make x) y)
	 | X2 x, X2 y ->
	     List.fold_left
	       (fun solved (p,v) ->
		  match X2.is_mine_r v with
		    | No t when X1.is_mine_t p && X1.is_mine_t t ->
			solve_list solved (solve1 p t)
		    | _ -> add_mt p (color2 v) solved)
	       solved (X2.solve x y)
	 | Empty x, X2 y | X2 y, Empty x -> 
	     List.fold_left
	       (fun solved (p,v) ->
		  match X2.is_mine_r v with
		    | No t when X1.is_mine_t p && X1.is_mine_t t ->
			solve_list solved (solve1 p t)
		    | _ -> add_mt p (color2 v) solved)
	       solved (X2.solve (X2.make x) y)
	 | _ -> raise Unsolvable)
      solved l

  let rec prodl acc = function
      [] -> acc
    | x::l -> List.fold_left (fun acc y-> (x,y)::acc) (prodl acc l) l

  let rec solve_rec mt ab = 
    let mt = solve_list mt ab in
    let mt , ab = 
      MT.fold 
	(fun t l ((mt,ab) as acc) -> match l with
	     _::_::_ -> MT.remove t mt , prodl ab l 
	   | _ -> acc)	 
	mt (mt,[])
    in 
    if ab=[] then mt else solve_rec mt ab
    
  let solve a b = try 
    MT.fold 
      (fun t l ret -> 
	 match l with [r] -> ((t , color r)::ret) | _ -> assert false) 
      (solve_rec MT.empty [a,b]) []
  with X1.Unsolvable | X2.Unsolvable -> raise Unsolvable
    
  module XX = struct type t'=t type t=t' let compare = compare end
  module Map = Map.Make(XX)
  module Set = Set.Make(XX)
    
  module R =
  struct
    type elt = t
    type t = { r1: X1.R.t ; r2: X2.R.t}
    exception Inconsistent    
    let empty = {r1=X1.R.empty ; r2=X2.R.empty}
   
    let assume env sa repr cl = try
      let repr1,repr2 = make_repr repr in
      let env1, seq1 = X1.R.assume env.r1 sa repr1 cl in
      let env2, seq2 = X2.R.assume env.r2 sa repr2 cl in
      {r1 = env1; r2 = env2}, Atom.SetEq.union seq1 seq2
    with X1.R.Inconsistent | X2.R.Inconsistent -> raise Inconsistent
      
    let query a repr cl env = 
      let repr1,repr2 = make_repr repr in
      if X1.is_mine_a a
      then X1.R.query a repr1 cl env.r1 
      else X2.R.query a repr2 cl env.r2
  end
end
and X1 : Sig.THEORY = struct
  include Arith
end
and X2 : Sig.THEORY = struct
  include Pairs
end




module Make (X : Sig.THEORY) = struct    

  module Uf = Uf.Make(X)
  module SF = Formula.Set
  module T = Term
  module A = Atom
  module MT = Term.Map
  module ST = Term.Set
  module SA = A.Set
  module SAeq = A.SetEq

  module MX = struct 
    include Map.Make(X)
    let find k m = try find k m with Not_found -> (ST.empty,SA.empty)
  end
    
  type t = { 
    use : (ST.t * SA.t) MX.t ; 
    uf : Uf.t ;
    relation : X.R.t ;
    proof_map : (T.t * SF.t) MT.t;
  }

  exception Inconsistent

  module Print = struct

    let sterms fmt = ST.iter (fprintf fmt "%a " T.print)
    let satoms fmt = SA.iter (fprintf fmt "%a " A.print)
    let lrepr fmt = List.iter (fprintf fmt "%a " X.print)
    let use env = 
      fprintf fmt "[cc] >>>> use table@.";
      MX.iter 
	(fun t (st,sa) -> 
	   fprintf fmt "%a used by @.{%a} and {%a} @."  
	     X.print t sterms st satoms sa) 
	env.use;
      fprintf fmt "[cc] <<<< @."

    let close_up t1 t2 = 
      fprintf fmt "[cc] close_up: %a=%a@." T.print t1 T.print t2
	
    let congruent t1 t2 = 
      fprintf fmt "[cc] cong %a=%a ?@." T.print t1 T.print t2

    let add_to_use t = fprintf fmt "[cc] add_to_use: %a@." T.print t

    let fathers t ft = 
      fprintf fmt "[cc] fathers of %a@." T.print t; sterms fmt ft
	
    let leaves t lvs = 
      fprintf fmt "[cc] leaves of %a@.@." T.print t; lrepr fmt lvs
  end

  let prod f acc l1 l2 = 
    List.fold_left 
      (fun acc x -> 
	 List.fold_left (fun acc y -> (f x y)::acc) acc l2) acc l1

  let compat_leaves env f l1 l2 =
    try
      Some(
	List.fold_left2 
	  (fun ex x y -> 
	     if not (Uf.equal env.uf x y) then raise Exit
	     else SF.union (Uf.explain env.uf x y) ex
	  ) SF.empty l1 l2 )
    with Exit -> None

  let congruent env u1 u2 = 
    if debug_cc then Print.congruent u1 u2;
    let {T.f=f1;xs=xs1} , {T.f=f2;xs=xs2} = T.view u1 , T.view u2 in
    if Symbols.equal f1 f2 then compat_leaves env f1 xs1 xs2 else None 
      
  let fathers t env = 
    X.Set.fold
      (fun x (st,sa) -> 
	 let st_x , sa_x = MX.find x env.use in
	 ST.union st st_x , SA.union sa sa_x)
      (Uf.leaves env.uf t) (ST.empty,SA.empty)

  let rec close_up t1 t2 proof env ccb =
    if debug_cc then Print.close_up t1 t2;
    if Uf.equal env.uf t1 t2 then env
    else
      let uf, res = Uf.union env.uf t1 t2 proof ccb in
      if Uf.Res.is_empty res then env 
      else
	let p,touched,leaves_P = Uf.Res.choose res in
	let (st_ftr_p, sa_ftr_p) as ftr_p = MX.find p env.use in
	let st_others, sa_others = 
	  ST.fold 
	    (fun t s -> 
	       X.Set.fold 
		 (fun x (st,sa) -> 
		    let st_x,sa_x = MX.find x env.use in
		    ST.union st st_x, SA.union sa sa_x) 
		 (Uf.leaves uf t) s)
	    touched ftr_p
	in
	let use = 
	  X.Set.fold 
	    (fun l use -> 
	       let st_l , sa_l = MX.find l env.use in
	       let st = ST.union st_l st_ftr_p in
	       let sa = SA.union sa_l sa_ftr_p in
	       MX.add l (st,sa) use)
	    leaves_P env.use (* prevoir le menage de r dans use *) in
	let env = ST.fold 
	  (fun x e -> 
	     ST.fold 
	       (fun y e -> 
		  match congruent e x y with
		      None -> e
		    | Some ex -> close_up x y ex e true)
	       st_others e)
	  st_ftr_p {env with use=use ; uf=uf}
	in
	replay_atom (SA.union sa_ftr_p sa_others) proof env

  and replay_atom sa proof env = 
    (* ATTENTION, proof est mal utilise *)
    let rel , nsa  = 
      X.R.assume env.relation sa (Uf.find env.uf) (Uf.classof env.uf) in
    SAeq.fold 
      (fun (a,t1,t2) env -> 
	 let r1 = Uf.find env.uf t1 in
	 let r2 = Uf.find env.uf t2 in
	 let st_r1 , sa_r1 = MX.find r1 env.use in
	 let st_r2 , sa_r2 = MX.find r2 env.use in
	 let sa_r1' , sa_r2' = SA.remove a sa_r1 , SA.remove a sa_r2 in
	 let use =  MX.add r1 (st_r1,sa_r1') env.use in
	 let use =  MX.add r2 (st_r2,sa_r2') use in	     
	 let env = close_up t1 t2 proof { env with use = use} true in
	 env
      ) nsa { env with relation = rel }
      
  let congruents e t s acc = 
    ST.fold 
      (fun t' acc -> 
	 match congruent e t t' with
	     None -> acc
	   | Some ex -> (t,t',ex)::acc) s acc
	      	
  let rec add_term (env,ct) t = 
    if debug_cc then Print.add_to_use t;
    let env = 
      if MT.mem t env.proof_map then env 
      else { env with proof_map = MT.add t (t,SF.empty) env.proof_map } in
    if Uf.mem t env.uf then (env,ct)
    else
      let env = { env with uf = Uf.add t env.uf} in
      let {T.xs=xs} = T.view t in
      let env , ct = List.fold_left add_term (env,ct) xs in
      let rt = Uf.find env.uf t in
      let env = 
	if MX.mem rt env.use then env 
	else { env with use = MX.add rt (ST.empty,SA.empty) env.use }
      in
      let lvs = List.fold_left 
	(fun s t -> X.Set.union s (Uf.leaves env.uf t)) X.Set.empty xs 
      in
      X.Set.fold
	(fun rx (env,ct) ->
	   let st_uset , sa_uset = MX.find rx env.use in
	   let ct = congruents env t st_uset ct in
	   {env with use=MX.add rx (ST.add t st_uset,sa_uset) env.use} , ct
	) lvs (env,ct) 
	
  let add a env =
    let st = A.terms_of a in
    let env = T.Set.fold
      (fun t env -> 
	 let env , ct = add_term (env,[]) t in
	 List.fold_right 
	   (fun (x,y,proof) e -> close_up x y proof e true) ct env) st env
    in 
    match A.view a with
	A.Eq _ | A.Neq _ -> env
      | _ ->
	  let lvs = T.Set.fold
	    (fun t lvs -> X.Set.union lvs (Uf.leaves env.uf t)) st X.Set.empty
	  in
	  X.Set.fold
	    (fun rx env ->
	       let st_uset , sa_uset = MX.find rx env.use in
	       {env with use=MX.add rx (st_uset,SA.add a sa_uset) env.use}
	    ) lvs env


  let assume a proof env = try
    if debug_cc then Print.use env;
    let env  = add a env in
    match A.view a with
	A.Eq(t1,t2) -> close_up t1 t2 proof env false
      | A.Neq(t1,t2)-> 
	  let env = { env with uf=Uf.distinct env.uf t1 t2 proof} in
	  replay_atom (SA.singleton a) proof env
      | _ -> replay_atom (SA.singleton a) proof env
  with Uf.Inconsistent | X.R.Inconsistent -> raise Inconsistent
	      
  let query a env = 
    if debug_cc then Print.use env;
    match A.view a with
	A.Eq(t1,t2) -> Uf.equal env.uf t1 t2
      | A.Neq(t1,t2) -> Uf.are_distinct env.uf t1 t2
      | _ -> 
	  X.R.query (A.neg a) (Uf.find env.uf) (Uf.classof env.uf) env.relation

  let explain a env = 
    match A.view a with
      | A.Eq(x,y) when T.equal x y -> SF.empty
      | A.Eq(x,y) when not(Uf.equal env.uf x y) -> 
	  SF.empty
      | A.Eq(x,y) -> Uf.explain env.uf x y
      | A.Neq(x,y) when T.equal x y -> SF.empty
      | A.Neq(x,y) when Uf.equal env.uf x y -> SF.empty
      | A.Neq(x,y) -> Uf.nexplain env.uf x y
      | _ -> assert false

  let classof env t = Uf.classof env.uf t
    
  let empty = 
    let env = { 
      use = MX.empty ; 
      uf = Uf.empty ; 
      relation = X.R.empty ;
      proof_map = MT.empty }
    in
    assume (A.make (A.Neq(T.vrai,T.faux))) SF.empty env 

  let normal_form = X.normal_form

end

